; Ngaro VM
; Copyright (c) 2010, Charles Childers
; -----------------------------------------------------------------------------
(defvar memory  (make-array 1000001) "memory for image"      )
(defvar stack   (make-array 1025)    "data stack"            )
(defvar address (make-array 1025)    "address stack"         )
(defvar ports   (make-array 1025)    "I/O ports"             )
(defvar ip      0                    "instruction pointer"   )
(defvar sp      0                    "data stack pointer"    )
(defvar rp      0                    "address stack pointer" )
(defvar i       0                    "scratch #1"            )
(defvar z       0                    "scratch #2"            )
(defvar opcode  0                    "current instruction"   )
(defvar in      0                    "input file pointer"    )
(defvar cells   0                    "size of image file"    )
; -----------------------------------------------------------------------------
(defun initialize-memory ()
  (setq i 0)
  (loop
    (if (> i 1000000) (return))
    (setf (aref memory i) 0)
    (setq i (+ i 1))))

(defun initialize-stacks ()
  (setq i 0)
  (loop
    (if (> i 1024) (return))
    (setf (aref stack i) 0)
    (setf (aref address i) 0)
    (setf (aref ports i) 0)
    (setq i (+ i 1))))

(defun bytes-to-int (b1 b2 b3 b4)
  (+ b1 (* b2 256) (* b3 65536) (* b4 16777216)))

(defun r ()
  (read-byte in))

(defun read-image-file (size)
  (setq i 0)
  (loop
    (if (= i size) (return))
    (setf (aref memory i)
      (bytes-to-int (r) (r) (r) (r)))
    (if (< 2147483647 (aref memory i))
                      (setf (aref memory i) (- (aref memory i) 4294967296)))
    (setq i (+ i 1))))

(defun load-image (filename)
  (setq in (open filename :element-type '(unsigned-byte 8)))
  (setq cells (/ (file-length in) 4))
  (read-image-file cells)
  (close in))

(defun handle-devices ()
(if (= (aref ports 0) 0)
(progn
  (cond
    ((= (aref ports 1) 1)
        (progn
          (setf (aref ports 1) (char-code (read-char *standard-input*)))
          (if (= (aref ports 1) 13) (progn (setf (aref ports 1) 10)))
          (setf (aref ports 0) 1)))
    ((= (aref ports 2) 1)
        (progn
          (princ (code-char (aref stack sp)))
          (setf (aref ports 2) 0)
          (setq sp (- sp 1))
          (setf (aref ports 0) 1)))
    ((= (aref ports 5) -1)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -2)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -3)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -4)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -5)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) sp)))
    ((= (aref ports 5) -6)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) rp)))
    ((= (aref ports 5) -7)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -8)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -9)
        (progn
          (setf (aref ports 0) 1)
          (setq ip 1000000)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -10)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -11)
        (progn
          (setf (aref ports 0) 1)
          (setf (aref ports 5) 0)))
    ((= (aref ports 5) -12)
        (progn
          (setf (aref ports 0) 1)  ;
          (setf (aref ports 5) 0)))))))

(defun process ()
  (setq opcode (aref memory ip))
  (cond
    ((= opcode 0)   ; nop
     ())
    ((= opcode 1)   ; lit
     (progn
       (setq sp (+ sp 1))
       (setq ip (+ ip 1))
       (setf (aref stack sp) (aref memory ip))))
    ((= opcode 2)   ; dup
     (progn
       (setq sp (+ sp 1))
       (setf (aref stack sp) (aref stack (- sp 1)))))
    ((= opcode 3)   ; drop
     (progn
       (setq sp (- sp 1))))
    ((= opcode 4)   ; swap
     (progn
       (setq i (aref stack sp))
       (setf (aref stack sp) (aref stack (- sp 1)))
       (setf (aref stack (- sp 1)) i)))
    ((= opcode 5)   ; push
     (progn
       (setq rp (+ rp 1))
       (setf (aref address rp) (aref stack sp))
       (setq sp (- sp 1))))
    ((= opcode 6)   ; pop
     (progn
       (setq sp (+ sp 1))
       (setf (aref stack sp) (aref address rp))
       (setq rp (- rp 1))))
    ((= opcode 7)   ; loop
     (progn
       (setf (aref stack sp) (- (aref stack sp) 1))
       (if (> (aref stack sp) 0)
         (progn
           (setq ip (+ ip 1))
           (setq ip (- (aref memory ip) 1)))
         (progn
           (setq ip (+ ip 1))
           (setq sp (- sp 1))))))
    ((= opcode 8)   ; jump
     (progn
       (setq ip (+ ip 1))
       (setq ip (- (aref memory ip) 1))
       (if (= (aref memory (+ ip 1)) 0) (setq ip (+ ip 1)))
       (if (= (aref memory (+ ip 1)) 0) (setq ip (+ ip 1)))))
    ((= opcode 9)   ; return
     (progn
       (setq ip (aref address rp))
       (setq rp (- rp 1))))
    ((= opcode 10)  ; > jump
     (progn
       (setq i (aref stack sp))
       (setq z (aref stack (- sp 1)))
       (setq ip (+ ip 1))
       (if (> z i) (setq ip (- (aref memory ip) 1)))
       (setq sp (- sp 2))))
    ((= opcode 11)  ; < jump
     (progn
       (setq i (aref stack sp))
       (setq z (aref stack (- sp 1)))
       (setq ip (+ ip 1))
       (if (< z i) (setq ip (- (aref memory ip) 1)))
       (setq sp (- sp 2))))
    ((= opcode 12)  ; != jump
     (progn
       (setq i (aref stack sp))
       (setq z (aref stack (- sp 1)))
       (setq ip (+ ip 1))
       (if (/= z i) (setq ip (- (aref memory ip) 1)))
       (setq sp (- sp 2))))
    ((= opcode 13)  ; == jump
     (progn
       (setq i (aref stack sp))
       (setq z (aref stack (- sp 1)))
       (setq ip (+ ip 1))
       (if (= z i) (setq ip (- (aref memory ip) 1)))
       (setq sp (- sp 2))))
    ((= opcode 14)  ; @
     (progn
       (setf (aref stack sp) (aref memory (aref stack sp)))))
    ((= opcode 15)  ; !
     (progn
       (setf (aref memory (aref stack sp)) (aref stack (- sp 1)))
       (setq sp (- sp 2))))
    ((= opcode 16)  ; +
     (progn
       (setf (aref stack (- sp 1)) (+ (aref stack (- sp 1)) (aref stack sp)))
       (setq sp (- sp 1))))
    ((= opcode 17)  ; -
     (progn
       (setf (aref stack (- sp 1)) (- (aref stack (- sp 1)) (aref stack sp)))
       (setq sp (- sp 1))))
    ((= opcode 18)  ; *
     (progn
       (setf (aref stack (- sp 1)) (* (aref stack (- sp 1)) (aref stack sp)))
       (setq sp (- sp 1))))
    ((= opcode 19)  ; /mod
     (progn
       (setq i (aref stack sp))
       (setq z (aref stack (- sp 1)))
       (setf (aref stack sp) (floor (/ z i)))
       (setf (aref stack (- sp 1)) (mod z i))))
    ((= opcode 20)  ; and
     (progn
       (setf (aref stack (- sp 1)) (logand (aref stack (- sp 1)) (aref stack sp)))
       (setq sp (- sp 1))))
    ((= opcode 21)  ; or
     (progn
       (setf (aref stack (- sp 1)) (logior (aref stack (- sp 1)) (aref stack sp)))
       (setq sp (- sp 1))))
    ((= opcode 22)  ; xor
     (progn
       (setf (aref stack (- sp 1)) (logxor (aref stack (- sp 1)) (aref stack sp)))
       (setq sp (- sp 1))))
    ((= opcode 23)  ; <<
     (progn
       (setq i (aref stack sp))
       (setq z (aref stack (- sp 1)))
       (setf (aref stack (- sp 1)) (ash z i))
       (setq sp (- sp 1))))
    ((= opcode 24)  ; >>
     (progn
       (setq i (aref stack sp))
       (setq z (aref stack (- sp 1)))
       (setq i (* i -1))
       (setf (aref stack (- sp 1)) (ash z i))
       (setq sp (- sp 1))))
    ((= opcode 25)  ; 0;
     (progn
       (if (= (aref stack sp) 0)
              (progn
                (setq ip (aref address rp))
                (setq rp (- rp 1))
                (setq sp (- sp 1))))))
    ((= opcode 26)  ; 1+
     (progn
       (setf (aref stack sp) (+ (aref stack sp) 1))))
    ((= opcode 27)  ; 1-
     (progn
       (setf (aref stack sp) (- (aref stack sp) 1))))
    ((= opcode 28)  ; in
     (progn
       (setf (aref stack sp) (aref ports (aref stack sp)))))
    ((= opcode 29)  ; out
     (progn
       (setf (aref ports (aref stack sp)) (aref stack (- sp 1)))
       (setq sp (- sp 2))))
    ((= opcode 30)  ; wait
     (handle-devices))
    ((> opcode 30)  ; call (implicit)
     (progn
       (setq rp (+ rp 1))
       (setf (aref address rp) ip)
       (setq ip (- opcode 1))
       (if (= (aref memory (+ ip 1)) 0) (setq ip (+ ip 1)))
       (if (= (aref memory (+ ip 1)) 0) (setq ip (+ ip 1)))))
  )
)

(defun ngaro ()
  (initialize-memory)
  (initialize-stacks)
  (load-image "retroImage")
  (loop
    (if (> ip 1000000) (return))
    (process)
    (setq ip (+ ip 1))))
; -----------------------------------------------------------------------------
(ngaro)
